# -*- coding: utf-8 -*-
from __future__ import absolute_import

from logging import getLogger
from migrate_tool import storage_service
from migrate_tool.task import Task
from qiniu import Auth
from qiniu import BucketManager
import requests

logger = getLogger(__name__)


class TokenException(Exception):
    pass


class QiniuStorageService(storage_service.StorageService):

    def __init__(self, *args, **kwargs):

        accesskeyid = kwargs['accesskeyid']
        accesskeysecret = kwargs['accesskeysecret']
        self._bucket = kwargs['bucket']
        self._auth = Auth(accesskeyid, accesskeysecret)
        self._domain = kwargs['domain_url']
        self._qiniu_api = BucketManager(self._auth)
        self._prefix = kwargs['prefix'] if 'prefix' in kwargs else ''

    def download(self, task, local_path):
        cos_path = task.key
        if isinstance(local_path, str):
            local_path = local_path.decode('utf-8')
        if cos_path.startswith('/'):
            cos_path = cos_path[1:]

        if isinstance(cos_path, unicode):
            cos_path = cos_path.encode('utf-8')
            from urllib import quote
            cos_path = quote(cos_path)

        base_url = 'http://%s/%s' % (self._domain, cos_path)
        # print base_url
        private_url = self._auth.private_download_url(base_url, expires=3600)
        # print private_url
        logger.debug("private url: " + private_url)

        for i in range(20):
            try:
                ret = requests.get(private_url)

                if ret.status_code != 200:
                    raise SystemError("download file from qiniu failed")

                with open(local_path.encode('utf-8'), 'wb') as fd:
                    for chunk in ret.iter_content(1024):
                        fd.write(chunk)
                if task.size is not None:
                    from os import path
                    if path.getsize(local_path.encode('utf-8')) != task.size:
                        raise IOError("Download failed error")
                    else:
                        logger.info("download successful")
                        break
            except:
                pass
        else:
            raise IOError("Download failed error")

    def upload(self, cos_path, local_path):
        raise NotImplementedError

    def list(self):
        limit = 100
        delimiter = None
        marker = None

        eof = False

        while not eof:
            try:
                ret, eof, info = self._qiniu_api.list(self._bucket, self._prefix, marker, limit, delimiter)
                if ret is None:
                    logger.warn("ret is None")
                    if info.error == 'bad token':
                        raise TokenException
                    else:
                        logger.warn(info.text_body)
                        raise IOError(info.error)

                for i in ret['items']:
                    logger.info("yield new object: {}".format(i['key']))
                    yield Task(i['key'], i['fsize'], None)

                if eof is True:
                    logger.info("eof is {}".format(eof))
                    continue

                if not eof and 'marker' in ret:
                    marker = ret['marker']
                else:
                    eof = True
            except TokenException as e:
                eof = True
                logger.warn("Your accessid/accesskey is incorrect, Please double check your configures")
            except Exception as e:
                logger.exception("list exception: " + str(e))

    def exists(self, _path):
        raise NotImplementedError



# -*- coding: utf-8 -*-


from logging import getLogger
from migrate_tool import storage_service
from qcloud_cos import CosClient

from qcloud_cos import UploadFileRequest, StatFileRequest, ListFolderRequest, DownloadFileRequest


from migrate_tool.task import Task


def to_unicode(s):
    if isinstance(s, str):
        return s.decode('utf-8')
    else:
        return s


def to_utf8(s):
    if isinstance(s, unicode):
        return s.encode('utf-8')
    else:
        return s


logger = getLogger(__name__)


class CosV4StorageService(storage_service.StorageService):

    def __init__(self, *args, **kwargs):

        appid = int(kwargs['appid'])
        region = kwargs['region']
        accesskeyid = unicode(kwargs['accesskeyid'])
        accesskeysecret = unicode(kwargs['accesskeysecret'])
        bucket = unicode(kwargs['bucket'])
        if 'prefix_dir' in kwargs:
            self._prefix_dir = kwargs['prefix_dir']
        else:
            self._prefix_dir = None

        self._cos_api = CosClient(appid, accesskeyid, accesskeysecret, region=region)
        self._bucket = bucket
        self._overwrite = kwargs['overwrite'] == 'true' if 'overwrite' in kwargs else False
        self._max_retry = 20

    def download(self, task, local_path):
        # self._oss_api.get_object_to_file(urllib.unquote(cos_path).encode('utf-8'), local_path)
        for i in range(20):
            logger.info("download file with rety {0}".format(i))
            import os
            try:
                os.remove(task.key)
            except:
                pass
            req = DownloadFileRequest(self._bucket, task.key, local_path)
            ret = self._cos_api.download_file(req)
            # self._oss_api.get_object_to_file(task.key, local_path)
            logger.debug(str(ret))

            if task.size is None:
                logger.info("task's size is None, skip check file size on local")
                break

            from os import path
            if path.getsize(local_path) != int(task.size):
                logger.error("Download Failed, size1: {size1}, size2: {size2}".format(size1=path.getsize(local_path),
                                                                                      size2=task.size))
            else:
                logger.info("Download Successfully, break")
                break
        else:
            raise IOError("Download Failed with 20 retry")

    def upload(self, task, local_path):
        cos_path = task.key
        if not cos_path.startswith('/'):
            cos_path = '/' + cos_path

        if self._prefix_dir:
            cos_path = self._prefix_dir + cos_path

        if isinstance(local_path, unicode):
            local_path.encode('utf-8')
        insert_only = 0 if self._overwrite else 1
        for i in range(10):
            try:
                upload_request = UploadFileRequest(self._bucket, unicode(cos_path), local_path, insert_only=insert_only)
                upload_file_ret = self._cos_api.upload_file(upload_request)

                if upload_file_ret[u'code'] != 0:
                    logger.warn("upload failed:" + str(upload_file_ret))
                    raise OSError("UploadError: " + str(upload_file_ret))
                else:
                    break
            except:
                pass
        else:
            raise IOError("upload failed")

    def list(self):
        if self._prefix_dir is None:
            for i in self.dfs('/'):
                yield i
        else:
            for i in self.dfs(self._prefix_dir):
                yield i

    def dfs(self, path):
        print "DFS: {path}".format(path=to_utf8(path))
        _finish = False
        _context = u''
        max_retry = self._max_retry

        path = to_unicode(path)

        while not _finish:
            request = ListFolderRequest(bucket_name=self._bucket, cos_path=path, context=_context)
            ret = self._cos_api.list_folder(request)

            if ret['code'] != 0:
                max_retry -= 1
            else:
                _finish = ret['data']['listover']
                _context = ret['data']['context']
                for item in ret['data']['infos']:
                    if 'filelen' in item:
                        try:
                            key = "{prefix}{filename}".format(prefix=path, filename=item['name'])
                            yield Task(key, item['filelen'], None)
                        except:
                            pass
                    else:
                        _sub_dir = "{prefix}{filename}".format(prefix=path.encode('utf-8'),
                                                               filename=item['name'].encode('utf-8'))
                        for i in self.dfs(_sub_dir):
                            yield i

            if max_retry == 0:
                _finish = True

    def exists(self, task):
        _path = task.key
        _size = task.size

        if not _path.startswith('/'):
            _path = '/' + _path
        logger = getLogger(__name__)
        # logger.info("func: exists: " + str(_path))
        if self._prefix_dir:
            _path = self._prefix_dir + _path

        if isinstance(_path, str):
            _path = _path.decode('utf-8')
        request = StatFileRequest(self._bucket, _path)
        ret = self._cos_api.stat_file(request)
        logger.debug("ret: " + str(ret))
        # import json
        # v = json.loads(ret)
        if ret['code'] != 0:
            # logger.warn("error code: " + str(ret['code']))
            return False
        if ret['data']['filelen'] != ret['data']['filesize']:
            logger.warn("file is broken, filelen: {len}, filesize: {size}".format(
                len=ret['data']['filelen'],
                size=ret['data']['filesize']
            ))
            return False
        elif _size is not None and ret['data']['filelen'] != _size:
            return False
        elif ret['data']['filelen'] != _size:
            return False
        return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os
from os import path
from logging import getLogger
from logging import getLogger, basicConfig, DEBUG
from sys import stderr
import time
from threading import Timer, Thread

from migrate_tool.worker import Worker
from migrate_tool.filter import Filter

logger = getLogger(__name__)


class BaseMigrator(object):

    def start(self):
        pass

    def stop(self):
        pass

    @property
    def status(self):
        """ Query migrate status

        :return: dict like {'success': 213, 'failure': 19, 'state': 'running'}
        """
        pass


class ThreadMigrator(BaseMigrator):
    """migrator Class, consisted of:
        1. Workers
        2. InputStorageService
        3. OutputStorageService
        4. Filter: Determines whether the file has been moved

    """

    def __init__(self, input_service, output_service, work_dir=None, threads=10, *args, **kwargs):

        self._input_service = input_service
        self._output_service = output_service

        self._work_dir = work_dir or os.getcwd()
        self._filter = Filter(self._work_dir)

        self._worker = Worker(work_dir=self._work_dir,
                              file_filter=self._filter,
                              input_service=self._input_service,
                              output_service=self._output_service,
                              threads_num=threads)

        self._stop = False
        self._finish = False
        self._threads = []

        # if path.exists(path.join(self._work_dir, 'filter.json')):
        #    with open(path.join(self._work_dir, 'filter.json'), 'r') as f:
        #       self._filter.loads(f.read())
        #        logger.info("loads bloom filter snapshot successfully.")

    def log_status_thread(self):
        while not self._stop:
            logger.info("working, {} tasks successfully, {} tasks failed.".format(self._worker.success_num,
                                                                                  self._worker.failure_num))
            time.sleep(3)

    def work_thread(self):
        assert self._output_service is not None
        try:
            for task in self._output_service.list():

                if self._stop:
                    break

                # print type(task)
                object_name_ = task.key
                if isinstance(object_name_, unicode):
                    object_name_ = object_name_.encode('utf-8')

                if self._filter.query(object_name_):
                    # object had been migrated
                    logger.info("{} has been migrated, skip it".format(object_name_))

                else:
                    # not migrated
                    self._worker.add_task(task)
                    logger.info("{} has been submitted, waiting for migrating".format(object_name_))
            else:
                self._finish = True
        except Exception as e:
            self._finish = True
            logger.exception(str(e))

    def start(self):
        log_status_thread = Thread(target=self.log_status_thread, name='log_status_thread')
        log_status_thread.daemon = True
        self._threads.append(log_status_thread)

        work_thread = Thread(target=self.work_thread, name='work_thread')
        work_thread.daemon = True
        self._threads.append(work_thread)

        for t in self._threads:
            t.start()

        self._worker.start()

    def stop(self, force=False):
        if force:
            self._worker.term()
        else:
            self._worker.stop()

        self._stop = True

        for t in self._threads:
            t.join()

    def status(self):
        return {'success': self._worker.success_num, 'fail': self._worker.failure_num, 'finish': self._finish}

if __name__ == '__main__':
    from migrate_tool.services.LocalFileSystem import LocalFileSystem

    migrator = ThreadMigrator(input_service=LocalFileSystem(workspace='F:\\Workspace\\tmp'),
                              output_service=LocalFileSystem(workspace='F:\\logstash-conf'))
    migrator.start()

    import time
    time.sleep(10)
    migrator.stop()



# -*- coding: utf-8 -*-

from logging import getLogger
from migrate_tool import storage_service
from migrate_tool.task import Task

from boto.s3.connection import S3Connection


logger = getLogger(__name__)


class S3StorageService(storage_service.StorageService):
    def __init__(self, *args, **kwargs):

        accesskeyid = kwargs['accesskeyid']
        accesskeysecret = kwargs['accesskeysecret']
        bucket = kwargs['bucket']
        self._prefix = kwargs['prefix'] if 'prefix' in kwargs else ''
        _s3_api = S3Connection(aws_access_key_id=accesskeyid, aws_secret_access_key=accesskeysecret)
        self._bucket_api = _s3_api.get_bucket(bucket)

    def download(self, task, local_path):
        for i in range(20):
            key = self._bucket_api.get_key(task.key)

            if key is not None:
                key.get_contents_to_filename(local_path)
            else:
                raise IOError("Download failed 404")
            if task.size is None:
                logger.info("task's size is None, skip check file size on local")
                break

            from os import path
            if path.getsize(local_path) != int(task.size):
                logger.error("Download Failed, size1: {size1}, size2: {size2}".format(size1=path.getsize(local_path),
                                                                                      size2=task.size))
            else:
                logger.info("Download Successfully, break")
                break
        else:
            raise IOError("download failed with 20 retry")

    def upload(self, cos_path, local_path):
        raise NotImplementedError

    def list(self):
        for obj in self._bucket_api.list(prefix=self._prefix):
            if obj.name[-1] == '/':
                continue
            logger.info("yield new object: {}".format(obj.key))
            yield Task(obj.name, obj.size, None)

    def exists(self, _path):
        raise NotImplementedError



# -*- coding: utf-8 -*-

from logging import getLogger
from migrate_tool import storage_service
from migrate_tool.task import Task

import oss2
logger = getLogger(__name__)


class OssStorageService(storage_service.StorageService):

    def __init__(self, *args, **kwargs):

        endpoint = kwargs['endpoint']
        accesskeyid = kwargs['accesskeyid']
        accesskeysecret = kwargs['accesskeysecret']
        bucket = kwargs['bucket']
        self._oss_api = oss2.Bucket(oss2.Auth(accesskeyid, accesskeysecret), endpoint, bucket)
        self._prefix = kwargs['prefix'] if 'prefix' in kwargs else ''
        if self._prefix.startswith('/'):
            self._prefix = self._prefix[1:]

    def download(self, task, local_path):
        # self._oss_api.get_object_to_file(urllib.unquote(cos_path).encode('utf-8'), local_path)
        for i in range(20):
            logger.info("download file with rety {0}".format(i))
            import os
            try:
                os.remove(task.key)
            except:
                pass

            self._oss_api.get_object_to_file(task.key, local_path)
            if task.size is None:
                logger.info("task's size is None, skip check file size on local")
                break

            from os import path
            if path.getsize(local_path) != int(task.size):
                logger.error("Download Failed, size1: {size1}, size2: {size2}".format(size1=path.getsize(local_path),
                                                                                      size2=task.size))
            else:
                logger.info("Download Successfully, break")
                break
        else:
            raise IOError("Download Failed with 20 retry")

    def upload(self, cos_path, local_path):
        raise NotImplementedError

    def list(self):
        for obj in oss2.ObjectIterator(self._oss_api, prefix=self._prefix):
            if obj.key[-1] == '/':
                continue
            logger.info("yield new object: {}".format(obj.key))
            yield Task(obj.key, obj.size, None)

    def exists(self, _path):
        raise NotImplementedError



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, with_statement
import os
from os import path
from migrate_tool.task import Task
from migrate_tool import storage_service


class LocalFileSystem(storage_service.StorageService):

    def __init__(self, *args, **kwargs):
        self._workspace = kwargs['workspace']

    def exists(self, path_):
        rt = path.join(self._workspace, path_)
        return path.exists(rt)

    def download(self, task, localpath):
        path_ = task['key']
        src_path = path.join(self._workspace, path_)
        import shutil
        return shutil.copyfile(src_path, localpath)

    def upload(self, task, localpath):
        path_ = task['key']
        src_path = path.join(self._workspace, path_)
        try:
            import os
            os.makedirs(path.dirname(src_path))
        except OSError:
            pass

        import shutil
        return shutil.copyfile(localpath, src_path)

    def list(self):
        for file in os.listdir(self._workspace):
            from os import path
            yield Task(file, path.getsize(file), None)


def make():
    """ hook function for entrypoints

    :return:
    """
    return LocalFileSystem

if __name__ == "__main__":
    import os
    fs = LocalFileSystem(workspace=os.getcwd())
    for f in fs.list():
        print(f)



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod


class StorageService(object):
    """ The abstract class for Storage Services. you must impl following functions.

    `path` is `/path/to/your/object`
    `localpath` is full local path.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def download(self, path, localpath):
        """ downloads object from service, and saves to local disk

        :param path: path on Services
        :param localpath: local path on Disk
        :return: success or failure
        """
        pass

    @abstractmethod
    def upload(self, path, localpath):
        """ uploads local file to service

        :param path: path on Service
        :param localpath: local path on Disk
        :return: success or failure
        """
        pass

    @abstractmethod
    def exists(self, path):
        """ query for existence of object

        :param path: path on Service
        :return:
        """
        pass

    @abstractmethod
    def list(self):
        """
        :return: iterator for 'a/b/c.txt', 'a/b/d.txt', not starts with '/'
        """
        pass



# -*- coding: utf-8 -*-

import leveldb
from os import path
from time import time


class Filter(object):

    def __init__(self, work_dir):
        self._workdir = work_dir
        self._db = leveldb.LevelDB(path.join(self._workdir, 'leveldb'))

    def add(self, value):
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        self._db.Put(value, str(time()))

    def query(self, value):
        try:
            if isinstance(value, unicode):
                value = value.encode('utf-8')
            self._db.Get(value)
            return True
        except KeyError:
            return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import pkg_resources
from ConfigParser import SafeConfigParser
from logging import getLogger, basicConfig, DEBUG
from sys import stderr
from argparse import ArgumentParser
import os
from os import path

from migrate_tool.migrator import ThreadMigrator

import signal
from logging.config import dictConfig
from threading import Thread
import sys
reload(sys)
sys.setdefaultencoding('utf8')

log_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
        'error': {
            'format': '%(asctime)s\t%(message)s'
        }
    },
    'handlers': {
        'default': {
            'level': 'DEBUG',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
        'error_file': {
            'level': 'DEBUG',
            'formatter': 'error',
            'class': 'logging.FileHandler',
            'filename': 'fail_files.txt',
            'mode': 'a'
        }
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': True
        },
        'migrate_tool.fail_file': {
            'handlers': ['error_file'],
            'level': 'DEBUG',
            'propagate': False
        },
        'requests.packages': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': True
        }
    }
}


services_ = {}


def loads_services():
    global services_
    for ep in pkg_resources.iter_entry_points(group='storage_services'):
        services_.update({ep.name: ep.load()})


def create_parser():
    parser_ = ArgumentParser()
    parser_.add_argument('-c', '--conf', type=file, required=True, help="specify your config")
    return parser_


def main_thread():

    parser = create_parser()
    opt = parser.parse_args()
    conf = SafeConfigParser()
    conf.readfp(opt.conf)

    output_service_conf = dict(conf.items('source'))
    input_service_conf = dict(conf.items('destination'))
    if conf.has_option('common', 'threads'):
        _threads = conf.getint('common', 'threads')
    else:
        _threads = 10
    workspace_ = conf.get('common', 'workspace')
    try:
        os.makedirs(workspace_)
    except OSError:
        pass

    log_config['handlers']['error_file']['filename'] = path.join(workspace_, 'failed_files.txt')
    dictConfig(log_config)

    loads_services()
    output_service = services_[output_service_conf['type']](**output_service_conf)
    input_service = services_[input_service_conf['type']](**input_service_conf)

    migrator = ThreadMigrator(input_service=input_service,
                              output_service=output_service,
                              work_dir=conf.get('common', 'workspace'),
                              threads=_threads)
    migrator.start()

    import time
    try:
        while True:
            state = migrator.status()

            if state['finish']:
                break
            time.sleep(3)

    except KeyboardInterrupt:
        state = migrator.status()
        print state
        import sys
        sys.exit()

    migrator.stop()
    state = migrator.status()
    print 'summary:\n ', 'failed: ', state['fail'], ' success: ', state['success']


def main_():
    thread_ = Thread(target=main_thread)
    thread_.daemon = True
    thread_.start()
    try:
        while thread_.is_alive():
            thread_.join(2)
    except KeyboardInterrupt:
        print 'exiting'


if __name__ == '__main__':
    main_()



#!/usr/bin/env python

from hashlib import sha256
import json


class BloomFilter(object):
    """A simple bloom filter"""

    def __init__(self, array_size=(1 * 1024), hashes=13):
        """Initializes a Filter() object
        Expects:
          array_size (in bytes): 4 * 1024 for a 4KB filter
          hashes (int): for the number of hashes to perform
        """
        self.filter = bytearray(array_size)  # The filter itself
        self.bitcount = array_size * 8  # Bits in the filter
        self.hashes = hashes  # The number of hashes to use

    def _hash(self, value):
        """Creates a hash of an int and yields a generator of hash functions
        Expects:
          value: int()
        Yields:
          generator of ints()
        """
        # Build an int() around the sha256 digest of int() -> value
        digest = int(sha256(value.__str__()).hexdigest(), 16)
        for _ in range(self.hashes):
            # bitwise AND of the digest and all of the available bit positions
            # in the filter
            yield digest & (self.bitcount - 1)
            # Shift bits in digest to the right, based on 256 (in sha256)
            # divided by the number of hashes needed be produced.
            # Rounding the result by using int().
            # So: digest >>= (256 / 13) would shift 19 bits to the right.
            digest >>= (256 / self.hashes)

    def add(self, value):
        """Bitwise OR to add value(s) into the self.filter
        Expects:
          value: generator of digest ints()
        """
        for digest in self._hash(value):
            # In-place bitwise OR of the filter, position is determined
            # by the (digest / 8) digest is described above in self._hash()
            # Bitwise OR is undertaken on the value at the location and
            # 2 to the power of digest modulo 8. Ex: 2 ** (30034 % 8)
            # to grantee the value is <= 128, the bytearray not being able
            # to store a value >= 256. Q: Why not use ((modulo 9) -1) then?
            self.filter[(digest / 8)] |= (2 ** (digest % 8))
            # The purpose here is to spread out the hashes to create a unique
            # "fingerprint" with unique locations in the filter array,
            # rather than just a big long hash blob.

    def query(self, value):
        """Bitwise AND to query values in self.filter
        Expects:
          value: value to check filter against (assumed int())
        """
        # If all() hashes return True from a bitwise AND (the opposite
        # described above in self.add()) for each digest returned from
        # self._hash return True, else False
        return all(self.filter[(digest / 8)] & (2 ** (digest % 8))
                   for digest in self._hash(value))

    def loads(self, s):
        ret = json.loads(s)

        self.hashes = ret['hashes']
        self.bitcount = ret['bitcount']
        self.filter = bytearray(ret['filter'])

    def dumps(self):
        ret = dict()
        ret['hashes'] = self.hashes
        ret['bitcount'] = self.bitcount
        ret['filter'] = [x for x in self.filter]
        return json.dumps(ret)


if __name__ == "__main__":
    bf = BloomFilter()

    bf.add(1234)
    bf.add(40005)
    bf.add(1)
    bf.add('helloworld')
    bf.add('helloworldforevery')

    print("Filter size {0} bytes").format(bf.filter.__sizeof__())

    print bf.query(1)  # True
    print bf.query(40005)  # True
    print bf.query(123)  # False
    print bf.query('helloworld')
    print bf.query('hellomath')
    s = bf.dumps()

    bf2 = BloomFilter()
    bf2.loads(s)

    print bf2.query(1)
    print bf2.query(2)



# -*- coding: utf-8 -*-


from migrate_tool import storage_service
from migrate_tool import task
from logging import getLogger
import requests
import urlparse
import hashlib

logger = getLogger(__name__)


class UrlListService(storage_service.StorageService):

    def __init__(self, *args, **kwargs):
        self._url_list_file = kwargs['url_list_file']
        self._timeout = float(kwargs['timeout'])
        self._chunk_size = 1024
        self._validator_method = None
        if kwargs.has_key("validator"):
            self._validator_method = kwargs['validator']

    def download(self, task, local_path):

        url_path = task.other
        expected_crc = task.size # size stores the sha1 or md5 of file

        for i in range(5):
            validator = None
	    if self._validator_method:
		if self._validator_method == "sha1":
	            validator = hashlib.sha1()
		elif self._validator_method == "md5":
		    validator = hashlib.md5()
		else:
	            validator = None
        
            try:
                ret = requests.get(url_path, timeout=self._timeout)
                if ret.status_code == 200:
                    with open(local_path, 'wb') as fd:
                        for chunk in ret.iter_content(self._chunk_size):
                            if validator:
                                validator.update(chunk)
                            fd.write(chunk)
                        fd.flush()
                    # validate 
                    if validator:
                         actual_crc = validator.hexdigest()
                         actual_crc_upper = actual_crc.upper()
                         if actual_crc != expected_crc and actual_crc_upper != expected_crc:
                             logger.debug("{}".format(str({'expected_crc:' : expected_crc, 'actual_crc:': actual_crc})))
                             raise IOError("NOTICE: downloaded file content not valid")
                    break
                else:
                    # print "task: ", task
                    raise IOError("NOTICE: download failed")
            except:
                logger.exception("download failed")
        else:
            raise IOError("NOTICE: download failed with retry 5")

    def upload(self, task, local_path):
        raise NotImplementedError

    def list(self):
        with open(self._url_list_file, 'r') as f:
            for line in f:
                try:
                    field = line.split()
                    if len(field) < 1:
                        logger.warn("{} is invalid".format(line))
                        continue
                    check_value = None
                    url_path = None
                    if len(field) == 1:
                        url_path = field[0]
                    else:
                        check_value = field[0].strip()
                        url_path = field[1]
                    ret = urlparse.urlparse(url_path)
                    if ret.path == '':
                        logger.warn("{} is invalid, No path".format(line))
                    logger.info("yield new object: {}".format(str({'store_path': ret.path.strip(), 'url_path': url_path.strip()})))
                    yield task.Task(ret.path.strip()[1:], check_value, url_path.strip())

                except Exception:
                    logger.warn("{} is invalid".format(line))

    def exists(self, _path):
        raise NotImplementedError



# -*- coding:utf-8 -*-
from Queue import Queue, Empty
from threading import Thread
from os import path, makedirs
from logging import getLogger
from threading import Lock

logger = getLogger(__name__)
fail_logger = getLogger('migrate_tool.fail_file')


class Worker(object):
    def __init__(self, work_dir, file_filter, input_service, output_service, threads_num=5, max_size=30):
        self._input_service = input_service
        self._output_service = output_service
        self._filter = file_filter
        self._work_dir = work_dir

        self._threads_num = threads_num
        self._threads_pool = []
        self._queue = Queue(maxsize=max_size)
        self._stop = False
        self._succ = 0
        self._fail = 0
        self._lock = Lock()

    def __work_thread(self):

        while not self._stop:
            # logger.info("worker stop: " + str(self._stop))
            try:
                # logger.debug("try to get task")
                task = self._queue.get_nowait()
                # logger.debug("get task succeefully")
                self._queue.task_done()
            except Empty:
                logger.debug("Empty queue" + str(self._stop))
                if self._stop:
                    break
                else:
                    import time
                    time.sleep(1)
                    continue

            task_path = task.key

            if task_path.startswith('/'):
                task_path = task_path[1:]

            if isinstance(task_path, str):
                task_path = task_path.decode('utf-8')

            import uuid
            localpath = unicode(path.join(self._work_dir, uuid.uuid1().hex))

            try:
                try:
                    makedirs(path.dirname(localpath))
                except OSError as e:
                    # directory is exists
                    logger.debug(str(e))

                try:
                    ret = self._input_service.exists(task)
                    if ret:
                        logger.info("{file_path} exists".format(file_path=task_path.encode('utf-8')))
                        with self._lock:
                            self._succ += 1
                            self._filter.add(task_path)
                        continue
                except Exception as e:
                    logger.exception("exists failed")

                try:
                    self._output_service.download(task, localpath)
                except Exception as e:
                    logger.exception("download failed")
                    self._fail += 1
                    fail_logger.error(task_path)
                    continue

                try:
                    self._input_service.upload(task, localpath)
                except Exception:
                    logger.exception("upload {} failed".format(task_path.encode('utf-8')))
                    self._fail += 1
                    fail_logger.error(task_path)
                    continue

                try:
                    import os
                    if isinstance(localpath, unicode):
                        localpath = localpath.encode('utf-8')

                    os.remove(localpath)
                    try:
                        os.removedirs(path.dirname(localpath))
                    except OSError:
                        pass
                except Exception as e:
                    logger.exception(str(e))
                    continue
                if isinstance(task_path, unicode):
                    logger.info("inc succ with {}".format(task_path.encode('utf-8')))
                else:
                    logger.info("inc succ with {}".format(task_path.encode('utf-8')))

                with self._lock:
                    self._succ += 1
                    self._filter.add(task_path)
            except Exception:
                logger.exception("try except for deleting file")

            finally:
                    import os
                    if isinstance(localpath, unicode):
                        localpath = localpath.encode('utf-8')

                    try:
                        os.remove(localpath)
                        os.removedirs(path.dirname(localpath))
                    except OSError:
                        pass

    def add_task(self, task):
        # blocking
        self._queue.put(task)

    def start(self):
        self._threads_pool = [Thread(target=self.__work_thread) for _ in range(self._threads_num)]
        for t in self._threads_pool:
            t.start()

    def stop(self):

        self._queue.join()
        self.term()

    def term(self):
        self._stop = True
        logger.info("try to stop migrate process.")
        # while any([t.is_alive() for t in self._threads_pool]):
        #     map(lambda i: i.join(5), filter(lambda j: j.is_alive(), self._threads_pool))
        #     print filter(lambda j: j.is_alive(), self._threads_pool)

        map(lambda i: i.join(), self._threads_pool)

    @property
    def success_num(self):
        return self._succ

    @property
    def failure_num(self):
        return self._fail



# -*- coding: utf-8 -*

from logging import getLogger
from migrate_tool import storage_service
from coscmd.client import CosS3Client, CosConfig, CosS3Auth

from migrate_tool.task import Task
import requests

def to_unicode(s):
    if isinstance(s, str):
        return s.decode('utf-8')
    else:
        return s


def to_utf8(s):
    if isinstance(s, unicode):
        return s.encode('utf-8')
    else:
        return s


logger = getLogger(__name__)

class CosS3StorageService(storage_service.StorageService):

    def __init__(self, *args, **kwargs):

        appid = int(kwargs['appid'])
        region = kwargs['region']
        accesskeyid = unicode(kwargs['accesskeyid'])
        accesskeysecret = unicode(kwargs['accesskeysecret'])
        bucket = unicode(kwargs['bucket'])
        if 'prefix_dir' in kwargs:
            self._prefix_dir = kwargs['prefix_dir']
        else:
            self._prefix_dir = None

        if 'part_size' in kwargs:
            self._part_size = kwargs['part_size']
        else:
            self._part_size = 1

        conf = CosConfig(appid=appid, access_id=str(accesskeyid), access_key=str(accesskeysecret), region=region, bucket=bucket, part_size=self._part_size)
        self._cos_client = CosS3Client(conf)
        self._bucket = bucket
        self._overwrite = kwargs['overwrite'] == 'true' if 'overwrite' in kwargs else False
        self._max_retry = 20
        self._appid = appid
        self._region = region
        self._accesskeyid = str(accesskeyid)
        self._accesskeysecret = str(accesskeysecret)

    def download(self, task, local_path):
        # self._oss_api.get_object_to_file(urllib.unquote(cos_path).encode('utf-8'), local_path)
        raise NotImplementedError

    def upload(self, task, local_path):
        cos_path = task.key
        if not cos_path.startswith('/'):
            cos_path = '/' + cos_path

        if self._prefix_dir:
            cos_path = self._prefix_dir + cos_path

        if isinstance(local_path, unicode):
            local_path.encode('utf-8')
        if cos_path.startswith('/'):
            cos_path = cos_path[1:]

        insert_only = 0 if self._overwrite else 1
        mp = self._cos_client.multipart_upload_from_filename(local_path, cos_path)
        for i in range(10):
            rt = mp.init_mp() 
            if rt:
                break
            logger.warn("init multipart upload failed")
        else:
            raise IOError("upload failed")
                
        for i in range(10):
            rt = mp.upload_parts()
            if rt:
                break
            logger.warn("multipart upload part failed")
        else:
            raise IOError("upload failed")
                    
        for i in range(10):
            rt = mp.complete_mp()
            if rt:
                break
            logger.warn("finish upload part failed")
        else:
            raise IOError("upload failed")

    def list(self):
         raise NotImplementedError

    def exists(self, task):
        cos_path = task.key
        cos_path = 'http://' + str(self._bucket) + '-' + str(self._appid) + '.' + str(self._region) + '.myqcloud.com' + '/' + str(cos_path)
        logger.debug('{}'.format(str({'cos_path:': cos_path})))
        try:
            ret = requests.head(cos_path, timeout=10, auth=CosS3Auth(self._accesskeyid, self._accesskeysecret))
            if ret.status_code != 200:
                return False
            else:
                return True
        except:
            logger.exception("head cos file failed")
            return False
        return True



from setuptools  import setup
# from distutils.core import setup
from platform import python_version_tuple


def requirements():

    with open('requirements.txt', 'r') as fileobj:
        requirements = [line.strip() for line in fileobj]

        version = python_version_tuple()

        if version[0] == 2 and version[1] == 6:
            requirements.append("argparse==1.4.0")
        return requirements

def long_description():
    with open('README.rst', 'r') as fileobj:
        return fileobj.read()

setup(
    name='cos_migrate_tool_test',
    version='0.1.12',
    packages=['migrate_tool', 'migrate_tool.services'],
    url='https://www.qcloud.com/',
    license='MIT',
    author='liuchang',
    author_email='liuchang0812@gmail.com',
    description='migrate tool for object storage services',
    long_description=long_description(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'cos_migrate_tool=migrate_tool.main:main_'
        ],
        'storage_services': [
            'localfs=migrate_tool.services.LocalFileSystem:LocalFileSystem',
            'oss=migrate_tool.services.oss:OssStorageService',
            'qiniu=migrate_tool.services.qiniu:QiniuStorageService',
            'cosv4=migrate_tool.services.cosv4:CosV4StorageService',
            'cosv3=migrate_tool.services.cosv3:CosV3StorageService',
            'coss3=migrate_tool.services.coss3:CosS3StorageService',
            'url=migrate_tool.services.url_list:UrlListService',
            's3=migrate_tool.services.s3:S3StorageService',
        ]
    },
    install_requires=requirements()
)



# -*- coding: utf-8 -*-


from logging import getLogger
from contextlib import closing

from migrate_tool import storage_service
from migrate_tool.task import Task
from qcloud_cos_v3 import CosClient
from qcloud_cos_v3 import ListFolderRequest
from qcloud_cos_v3 import Auth, CredInfo

import requests

logger = getLogger(__name__)


def to_utf8(s):
    if isinstance(s, unicode):
        return s.encode('utf-8')
    else:
        return s


class CosV3StorageService(storage_service.StorageService):
    def __init__(self, *args, **kwargs):

        appid = int(kwargs['appid'])
        accesskeyid = unicode(kwargs['accesskeyid'])
        accesskeysecret = unicode(kwargs['accesskeysecret'])
        bucket = unicode(kwargs['bucket'])
        if 'prefix' in kwargs:
            self._prefix = kwargs['prefix']
        else:
            self._prefix = None

        self._cos_api = CosClient(appid, accesskeyid, accesskeysecret)
        self._bucket = bucket
        self._appid = appid
        self._http_session = requests.session()
        self._auth = Auth(CredInfo(appid, accesskeyid, accesskeysecret))

    def __sign(self, key):
        import time
        ex = int(time.time()) + 300

        return self._auth.sign_download(self._bucket, key, ex)

    def download(self, task, local_path):
        key = task.key
        url_tmpl = 'http://{bucket}-{appid}.cos.myqcloud.com/{key}?sign={sign}'
        import urllib
        uri = url_tmpl.format(bucket=self._bucket, appid=self._appid, key=urllib.quote(to_utf8(key)),
                              sign=self.__sign(key))
        session = self._http_session

        for i in range(5):
            try:
                with closing(session.get(uri, stream=True)) as ret:
                    if ret.status_code in [200, 206]:
                        with open(local_path, 'wb') as f:
                            for chunk in ret.iter_content(chunk_size=1024):
                                if chunk:
                                    f.write(chunk)
                            f.flush()
                    else:
                        raise IOError("download failed " + ret.text)
                from os import path
                if task.size is not None and path.getsize(local_path) != task.size:
                    raise IOError("download failed with retry {times}".format(times=i))
                else:
                    break

            except Exception:
                logger.exception("download failed with retry {times}".format(times=i))
                import time
                time.sleep(2 ** min(5, i))
        else:
            raise IOError("download failed!")

    def upload(self, task, local_path):
        raise NotImplementedError

    def __dfs_list(self, path):
        logger.info("try to dump file list under {}".format(path))

        _finish = False
        _context = u''
        max_retry = 10
        while not _finish:
            try:
                request = ListFolderRequest(bucket_name=self._bucket, cos_path=path, context=_context)
                ret = self._cos_api.list_folder(request)
            except Exception:
                logger.exception("list failed")
                max_retry -= 1
                continue

            logger.debug(str(ret))

            if ret['code'] != 0:
                logger.warning("request failed: {}".format(str(ret)))
                max_retry -= 1
            else:
                _finish = not ret['data']['has_more']
                _context = ret['data']['context']
                for item in ret['data']['infos']:
                    if 'filelen' in item:
                        # file
                        key = "{prefix}{filename}".format(prefix=path, filename=item['name'])
                        yield Task(key, item['filelen'], None)
                    else:
                        _sub_dir = "{prefix}{filename}/".format(prefix=path, filename=item['name'])
                        if isinstance(_sub_dir, str):
                            _sub_dir = _sub_dir.decode('utf-8')
                        for i in self.__dfs_list(_sub_dir):
                            yield i
                            # directory

            if max_retry == 0:
                _finish = True
                logger.error("reach max retry times, finish this directory {}".format(path))

        logger.info("finish directory {}".format(path))

    def list(self):
        if self._prefix is None:
            prefix = u'/'
        else:
            prefix = unicode(self._prefix)

        for i in self.__dfs_list(prefix):
            yield i

    def exists(self, task):
        raise NotImplementedError
